#  Copyright 2020 Cognite AS
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import List, Optional

import termcolor

try:
    import readline  # noqa
except ModuleNotFoundError:
    pass


def prompt(name: str, default: Optional[str] = None) -> str:
    prompt_default = f" [default: {default}]" if default else ""

    while True:
        answer = input(f"Enter {name}{prompt_default}: ")

        if not answer:
            if default:
                return default
        else:
            return answer


def choice(name: str, choices: List[str], default: Optional[str] = None) -> str:
    while True:
        answer = prompt(f"{name} ({'/'.join(choices)})", default)
        if answer in choices:
            return answer


def headerprint(s: str) -> None:
    termcolor.cprint(s, "green", attrs=["bold"])


def errorprint(s: str) -> None:
    print(f"{termcolor.colored('Error:', 'red', attrs=['bold'])} {termcolor.colored(s, 'red')}")


def lineprint(color: str = "cyan") -> None:
    print()
    termcolor.cprint("-" * 79, color)
    print()


def get_git_user() -> Optional[str]:
    name = os.popen("git config --get user.name").read().strip()
    email = os.popen("git config --get user.email").read().strip()

    if name and email:
        return f"{name} <{email}>"
    return None



#  Copyright 2021 Cognite AS
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import Dict

import requests

from cognite.extractorutils.cogex.io import choice, get_git_user, prompt

pyproject_template = """
[tool.poetry]
name = "{name}"
version = "1.0.0"
description = "{description}"
authors = ["{author}"]

[tool.black]
line-length = 120
target_version = ['py38']
include = '\\.py$'

[tool.isort]
line_length=120                # corresponds to -w  flag
multi_line_output=3            # corresponds to -m  flag
include_trailing_comma=true    # corresponds to -tc flag
skip_glob = '^((?!py$).)*$'    # this makes sort all Python files
known_third_party = ["cognite"]

[tool.poetry.dependencies]
python = ">=3.8,<3.12"

[tool.poetry.dev-dependencies]
pyinstaller = "^5.12"
macholib = {{version = "^1.14", platform = "darwin"}}             # Used by pyinstaller pn Mac OS
pywin32-ctypes = {{version = "^0.2.0", platform = "win32"}}       # Used by pyinstaller on Windows
pefile = "^2022.5.30"                                           # Used by pyinstaller on Windows

[tool.poetry.scripts]
{name} = "{name}.__main__:main"

[build-system]
requires = ["poetry>=0.12"]
build-backend = "poetry.masonry.api"
"""

pre_commit_template = """
repos:
-   repo: https://github.com/psf/black
    rev: 23.7.0
    hooks:
      - id: black
-   repo: https://github.com/charliermarsh/ruff-pre-commit
    hooks:
    -   id: ruff
        args: [--fix, --exit-non-zero-on-fix]
    rev: v0.0.282
-   repo: local
    hooks:
      - id: mypy
        name: mypy
        entry: poetry run mypy --non-interactive --install-types
        language: system
        types: [python]
        pass_filenames: true
"""

mypy_config_template = """
[mypy]
disallow_untyped_defs = true
ignore_missing_imports = true

[mypy-tests.*]
ignore_errors = True
"""

basic_config_template = """
from dataclasses import dataclass

from cognite.extractorutils.configtools import BaseConfig, StateStoreConfig


@dataclass
class ExtractorConfig:
    state_store: StateStoreConfig = StateStoreConfig()


@dataclass
class Config(BaseConfig):
    extractor: ExtractorConfig = ExtractorConfig()
"""

basic_configfile_template = """
logger:
    console:
        level: INFO

cognite:
    # Read these from environment variables
    host: ${{COGNITE_BASE_URL}}
    project: ${{COGNITE_PROJECT}}

    idp-authentication:
        token-url: ${{COGNITE_TOKEN_URL}}

        client-id: ${{COGNITE_CLIENT_ID}}
        secret: ${{COGNITE_CLIENT_SECRET}}
        scopes:
            - ${{COGNITE_BASE_URL}}/.default
"""

basic_main_template = """
from cognite.extractorutils import Extractor

from {name} import __version__
from {name}.extractor import run_extractor
from {name}.config import Config


def main() -> None:
    with Extractor(
        name="{name}",
        description="{description}",
        config_class=Config,
        run_handle=run_extractor,
        version=__version__,
    ) as extractor:
        extractor.run()


if __name__ == "__main__":
    main()
"""

basic_extractor_template = """
from threading import Event

from cognite.client import CogniteClient
from cognite.extractorutils.statestore import AbstractStateStore

from {name}.config import Config


def run_extractor(cognite: CogniteClient, states: AbstractStateStore, config: Config, stop_event: Event) -> None:
    print("Hello, world!")
"""

basic_template: Dict[str, str] = {
    "example_config.yaml": basic_configfile_template,
    os.path.join("{name}", "config.py"): basic_config_template,
    os.path.join("{name}", "extractor.py"): basic_extractor_template,
    os.path.join("{name}", "__main__.py"): basic_main_template,
}


rest_configfile_template = """
logger:
    console:
        level: INFO

cognite:
    # Read these from environment variables
    host: ${{COGNITE_BASE_URL}}
    project: ${{COGNITE_PROJECT}}

    idp-authentication:
        token-url: ${{COGNITE_TOKEN_URL}}

        client-id: ${{COGNITE_CLIENT_ID}}
        secret: ${{COGNITE_CLIENT_SECRET}}
        scopes:
            - ${{COGNITE_BASE_URL}}/.default

source:
    auth:
        basic:
            username: my-user
            password: ${{SOURCE_PASSWORD}}
"""

rest_main_template = """
from {name} import __version__
from {name}.extractor import extractor


def main() -> None:
    with extractor:
        extractor.run()


if __name__ == "__main__":
    main()
"""

rest_extractor_template = """
import arrow

from cognite.extractorutils.uploader_types import Event
from cognite.extractorutils.rest.extractor import RestExtractor

from {name} import __version__
from {name}.dto import ElementList

extractor = RestExtractor(
    name="{name}",
    description="{description}",
    version=__version__,
)

@extractor.get("/elements", response_type=ElementList, interval=60)
def elements_to_events(element_list: ElementList) -> Iterable[Event]:
    for element in element_list.elements:
        start_time = arrow.get(element.startTime)
        end_time = start_time.shift(minutes=element.duration)

        yield Event(start_time=start_time, end_time=end_time, description=element.description)
"""

rest_dto_template = """
from dataclasses import dataclass
from typing import List

@dataclass
class MyElement:
    elementId: str
    eventTime: str
    duration: int
    description: str

@dataclass
class ElementList:
    elements: List[MyElement]
"""

mqtt_configfile_template = """
logger:
    console:
        level: INFO

cognite:
    # Read these from environment variables
    host: ${{COGNITE_BASE_URL}}
    project: ${{COGNITE_PROJECT}}

    idp-authentication:
        token-url: ${{COGNITE_TOKEN_URL}}

        client-id: ${{COGNITE_CLIENT_ID}}
        secret: ${{COGNITE_CLIENT_SECRET}}
        scopes:
            - ${{COGNITE_BASE_URL}}/.default

source:
    username: mqtt-user
    password: ${{MQTT_PASSWORD}}
    client-id: my-extractor
    host: mqtt://some-mqtt-broker
"""

mqtt_main_template = """
from {name}.extractor import extractor

def main() -> None:
    with extractor:
        extractor.run()

if __name__ == "__main__":
    main()
"""

mqtt_extractor_template = """
import arrow

from cognite.extractorutils.uploader_types import Event
from cognite.extractorutils.mqtt.extractor import MqttExtractor

from {name} import __version__
from {name}.dto import EventFromMqtt

extractor = MqttExtractor(
    name="{name}",
    description="{description}",
    version=__version__,
)

@extractor.topic(topic="mytopic", qos=1, response_type=EventFromMqtt)
def subscribe_events(evt: EventFromMqtt) -> Event:
    start_time = arrow.get(evt.startTime)
    end_time = start_time.shift(minutes=evt.duration)

    return Event(start_time=start_time, end_time=end_time, description=evt.description)
"""

mqtt_dto_template = """
from dataclasses import dataclass

@dataclass
class EventFromMqtt:
    elementId: str
    eventTime: str
    duration: int
    description: str
"""


rest_template: Dict[str, str] = {
    "example_config.yaml": rest_configfile_template,
    os.path.join("{name}", "extractor.py"): rest_extractor_template,
    os.path.join("{name}", "dto.py"): rest_dto_template,
    os.path.join("{name}", "__main__.py"): rest_main_template,
}

mqtt_template: Dict[str, str] = {
    "example_config.yaml": mqtt_configfile_template,
    os.path.join("{name}", "extractor.py"): mqtt_extractor_template,
    os.path.join("{name}", "dto.py"): mqtt_dto_template,
    os.path.join("{name}", "__main__.py"): mqtt_main_template,
}

templates: Dict[str, Dict[str, str]] = {"basic": basic_template, "rest": rest_template, "mqtt": mqtt_template}


def initialize_project() -> None:
    name = prompt("extractor name").replace(" ", "_").replace("-", "_").lower()
    description = prompt("description")
    author = prompt("author", get_git_user())

    print("Which template should be loaded?")
    print("  basic: loads a generic template suitable for most source systems")
    print(
        "  rest:  loads a template for extracting from RESTful APIs, using the REST\n"
        "         extension for extractor-utils"
    )
    print(
        "  mqtt:  loads a template for extracting data published by MQTT brokers, using\n"
        "         the MQTT extension for extractor-utils"
    )
    template = choice("template", ["basic", "rest", "mqtt"], "basic")

    with open("pyproject.toml", "w") as pyproject_file:
        pyproject_file.write(pyproject_template.format(name=name, description=description, author=author))
    with open("mypy.ini", "w") as mypy_file:
        mypy_file.write(mypy_config_template)
    with open(".pre-commit-config.yaml", "w") as pre_commit_file:
        pre_commit_file.write(pre_commit_template)
    print("Fetching gitignore template from GitHub")
    gitignore_template = requests.get("https://raw.githubusercontent.com/github/gitignore/master/Python.gitignore").text
    with open(".gitignore", "w") as gitignore_file:
        gitignore_file.write(gitignore_template)

    if not os.path.isdir(".git"):
        os.system("git init")

    os.mkdir(name)
    with open(os.path.join(name, "__init__.py"), "w") as init_file:
        init_file.write('__version__ = "1.0.0"')
    for path, content in templates[template].items():
        with open(path.format(name=name), "w") as f:
            f.write(content.format(name=name, description=description))

    os.system("poetry run pip install --upgrade pip")
    os.system("poetry add cognite-extractor-utils")
    if template == "rest":
        os.system("poetry add cognite-extractor-utils-rest")
    elif template == "mqtt":
        os.system("poetry add cognite-extractor-utils-mqtt")
    os.system("poetry add -D mypy flake8 black isort pre-commit")
    os.system("poetry lock")
    os.system("poetry install")
    os.system("poetry run pre-commit autoupdate")
    os.system("poetry run pre-commit install")
    os.system(f"poetry run black {name}")
    os.system(f"poetry run isort {name}")



#  Copyright 2020 Cognite AS
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import os
import sys
from argparse import ArgumentParser

from cognite.extractorutils.cogex.build import build, generate_spec
from cognite.extractorutils.cogex.docker import build_docker_image, create_dockerfile
from cognite.extractorutils.cogex.initialize import initialize_project
from cognite.extractorutils.cogex.io import errorprint


def get_argparser() -> ArgumentParser:
    argparser = ArgumentParser("cogex", description="A project manager for Cognite extractors")
    argparser.set_defaults(command=None)
    subparsers = argparser.add_subparsers(help="Command to run")

    build_parser = subparsers.add_parser(
        "build", help="Build artifacts", description="Build self-contained executables of Python extractors"
    )
    build_parser.add_argument("--spec", nargs="+", help="Build from spec file(s)", metavar="SPECFILE", default=None)
    build_parser.add_argument(
        "--dockerfile", help="Build docker file instead of native executable", action="store_true"
    )
    build_parser.add_argument(
        "--dockerimage", help="Build docker image instead of native executable", action="store_true"
    )
    build_parser.set_defaults(command="build")

    init_parser = subparsers.add_parser(
        "init", help="Initialize project directory", description="Create a new extractor project in this directory"
    )
    init_parser.set_defaults(command="init")

    spec_parser = subparsers.add_parser(
        "generate-spec",
        help="Generate spec file(s)",
        description="Generate spec file(s) for the project based on the script(s) in pyproject.toml",
    )
    spec_parser.set_defaults(command="spec")

    return argparser


def main() -> None:
    argparser = get_argparser()
    args = argparser.parse_args()

    try:
        if args.command == "build":
            try:
                os.mkdir("build")
            except FileExistsError:
                pass
            if args.dockerfile:
                create_dockerfile()
            elif args.dockerimage:
                build_docker_image()
            else:
                build(args)

        elif args.command == "spec":
            generate_spec()

        elif args.command == "init":
            initialize_project()

        else:
            argparser.print_usage()
            errorprint("No command given")
            sys.exit(1)

    except ValueError as e:
        errorprint(str(e))
        sys.exit(1)


if __name__ == "__main__":
    main()



from typing import Any, MutableMapping

import toml

_pyproject_singleton = None


def get_pyproject() -> MutableMapping[str, Any]:
    global _pyproject_singleton

    if _pyproject_singleton is None:
        _pyproject_singleton = toml.load("pyproject.toml")

    return _pyproject_singleton



import os
import subprocess
import sys
from time import time
from typing import List, Tuple

from cognite.extractorutils.cogex._common import get_pyproject
from cognite.extractorutils.cogex.io import errorprint, headerprint, lineprint

_dockerfile_template = """
FROM {docker_base}
{preamble}
RUN set -ex && pip install --upgrade pip && pip install poetry pipx
ENV PATH /root/.local/bin:$PATH

RUN mkdir -p {installdir}
WORKDIR {installdir}

COPY pyproject.toml ./
COPY poetry.lock ./
{copy_packages}

RUN pipx install .

RUN mkdir -p /config
COPY build/config_remote.yaml /config/config_remote.yaml

WORKDIR {workdir}

ENTRYPOINT [ "{entrypoint}" ]
CMD ["/config/config_remote.yaml"]
"""

_remote_configfile = """
type: remote
cognite:
    # Read these from environment variables
    host: ${COGNITE_BASE_URL}
    project: ${COGNITE_PROJECT}

    idp-authentication:
        token-url: ${COGNITE_TOKEN_URL}

        client-id: ${COGNITE_CLIENT_ID}
        secret: ${COGNITE_CLIENT_SECRET}
        scopes:
            - ${COGNITE_BASE_URL}/.default

    extraction-pipeline:
        external-id: ${COGNITE_EXTRACTION_PIPELINE}
"""


def _get_python_version() -> Tuple[int, int, int]:
    raw_version = (
        subprocess.check_output(["poetry", "run", "python", "-V"]).decode("ascii").replace("Python", "").strip()
    )
    print(f"Detected python version {raw_version}")
    raw_parts = raw_version.split(".")
    return int(raw_parts[0]), int(raw_parts[1]), int(raw_parts[2])


def _get_entrypoint() -> str:
    pyproject = get_pyproject()
    scripts = pyproject["tool"]["poetry"].get("scripts", [])

    if len(scripts) == 0:
        raise ValueError("No scripts found in [tool.poetry.scripts], can't deduce entrypoint")
    elif len(scripts) > 1:
        try:
            entrypoint = pyproject["tool"]["cogex"]["docker"]["entrypoint"]
        except KeyError:
            raise ValueError(
                "Multiple scripts found in [tool.poetry.scripts], "
                "please specify which is the entrypoint in 'entrypoint' under [tool.cogex.docker]"
            )

        if entrypoint not in scripts:
            raise ValueError(f"Given entrypoint {entrypoint} is not listed under [tool.poetry.scripts]")

        return entrypoint
    else:
        entrypoint = list(scripts.keys())[0]

    print(f"Using entrypoint '{entrypoint}' ({scripts[entrypoint].split(':')[0]})")
    return entrypoint


def _get_packages() -> List[str]:
    pyproject = get_pyproject()

    try:
        packages = [
            os.path.join(package["from"], package["include"]) for package in pyproject["tool"]["poetry"]["packages"]
        ]
    except KeyError:
        packages = [pyproject["tool"]["poetry"]["name"].replace("-", "_")]

    return packages


def _get_docker_base() -> str:
    try:
        base = get_pyproject()["tool"]["cogex"]["docker"]["base-image"]

    except KeyError:
        python_version = _get_python_version()
        base = f"python:{python_version[0]}.{python_version[1]}-slim"

    print(f"Using base image {base}")
    return base


def create_dockerfile() -> None:
    headerprint("Generating Dockerfile")
    packages = _get_packages()
    pyproject = get_pyproject()

    try:
        dockerconfig = pyproject["tool"]["cogex"]["docker"]
    except KeyError:
        raise ValueError("No [tool.cogex.docker] section in pyproject")

    preamble = dockerconfig.get("preamble", "")
    if preamble:
        print("Including preamble")

    copy_statements = ["COPY {} {}".format(p, p) for p in packages]

    if "readme" in pyproject["tool"]["poetry"]:
        copy_statements.append(f"COPY {pyproject['tool']['poetry']['readme']} {pyproject['tool']['poetry']['readme']}")
    copy_packages = "\n".join(copy_statements)

    installdir = dockerconfig.get("install-dir", pyproject["tool"]["poetry"]["name"].replace(" ", "_"))
    workdir = dockerconfig.get("work-dir", installdir)

    with open(f"build{os.path.sep}Dockerfile", "w") as dockerfile:
        dockerfile.write(
            _dockerfile_template.format(
                docker_base=_get_docker_base(),
                preamble=preamble,
                workdir=workdir,
                installdir=installdir,
                copy_packages=copy_packages,
                entrypoint=_get_entrypoint(),
            ).lstrip()
        )
    print(f"Dockerfile created at build{os.path.sep}Dockerfile")

    with open(f"build{os.path.sep}config_remote.yaml", "w") as remote_config:
        remote_config.write(_remote_configfile)


def build_docker_image() -> None:
    start_time = time()

    pyproject = get_pyproject()
    try:
        tags = [
            tag.format(version=pyproject["tool"]["poetry"]["version"])
            for tag in get_pyproject()["tool"]["cogex"]["docker"]["tags"]
        ]
    except KeyError:
        raise ValueError("No docker tags listed in 'tags' under [tool.cogex.docker]")

    create_dockerfile()

    headerprint("Building Docker image")
    formatted_tags = " ".join([f"-t {tag}" for tag in tags])
    if os.system(f"docker build . -f build{os.path.sep}Dockerfile {formatted_tags}"):
        lineprint("red")
        errorprint("Build failed")
        print()
        sys.exit(1)

    lineprint()
    headerprint("Build done")
    print(f"Created docker images: {', '.join(tags)}")
    print(f"Total build time: {time() - start_time:.1f} s")
    print()



#  Copyright 2020 Cognite AS
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import fileinput
import os
import re
import sys
from argparse import Namespace
from time import time
from typing import Dict, List

from cognite.extractorutils.cogex._common import get_pyproject
from cognite.extractorutils.cogex.io import errorprint, headerprint, lineprint

_spec_template = """
# -*- mode: python ; coding: utf-8 -*-

import os
import sys

try:
    from {package} import __version__
    version = f"{{__version__}}-"
except ImportError:
    version = ""

if os.name == "nt":
    script_path = "{script_path_win}"

else:
    script_path = "{script_path_unix}"


a = Analysis(
    [script_path],
    binaries=[],
    datas=[],
    hiddenimports=[],
    hookspath=[],
    runtime_hooks=[],
    excludes=[],
    win_no_prefer_redirects=False,
    win_private_assemblies=False,
    cipher=None,
    noarchive=False,
)
pyz = PYZ(a.pure, a.zipped_data, cipher=None)
exe = EXE(
    pyz,
    a.scripts,
    a.binaries,
    a.zipfiles,
    a.datas,
    [],
    name=f"{name}-{{version}}{{sys.platform}}",
    debug=False,
    bootloader_ignore_signals=False,
    strip=False,
    upx=True,
    upx_exclude=[],
    runtime_tmpdir=None,
    console=True,
    #icon="logo.ico",
)
"""


def update_version(initfile: str) -> None:
    print(f"Updating {initfile}")

    init_version_search = re.search(r'^__version__\s*=\s*"(.*)"', open(initfile).read(), re.M)
    if init_version_search is None:
        print("No __version__ found")
        return

    init_version = init_version_search.group(1)
    print(f"Version in {initfile}: {init_version}")

    poetry_version = get_pyproject()["tool"]["poetry"]["version"]
    print(f"Version in pyproject.toml: {poetry_version}")

    if poetry_version == init_version:
        print("No updating necessary")
        return

    print("Changing __version__ to match pyproject")
    with fileinput.input(initfile, inplace=True) as file:
        for line in file:
            print(line.replace(init_version, poetry_version), end="")


def build(args: Namespace) -> None:
    start_time = time()

    pyproject = get_pyproject()

    # Ensure spec file exists
    if args.spec is None:
        # No spec file, attempt to create one on-the-fly
        headerprint("Generating spec files")
        os.makedirs("build", exist_ok=True)
        specfiles = generate_spec(keep_spec=False)

    else:
        specfiles = args.spec

    # Make sure __version__ is updated if present
    headerprint("Updating __version__ in __init__.py from pyproject.toml")
    try:
        packages = [
            os.path.join(package["from"], package["include"]) for package in pyproject["tool"]["poetry"]["packages"]
        ]
    except KeyError:
        packages = [pyproject["tool"]["poetry"]["name"]]

    for package in packages:
        for dirpath, dirnames, filenames in os.walk(package):
            for filename in filenames:
                if filename == "__init__.py":
                    update_version(os.path.join(dirpath, filename))

    headerprint("Making sure Poetry environment is up to date")
    os.system("poetry install")

    # Build executables
    s = "s" if len(specfiles) > 1 else ""
    headerprint(f"Building executable{s}")
    for specfile in specfiles:
        if os.system(f"poetry run pyinstaller {specfile}"):
            lineprint("red")
            errorprint(f"Build of {specfile} failed")
            print()
            sys.exit(1)

    lineprint()
    headerprint("Build done")

    print(f"Resulting file{s} found in ./dist")
    print(f"Total build time: {time() - start_time:.1f} s")
    print()


def generate_spec(keep_spec: bool = True) -> List[str]:
    pyproject = get_pyproject()

    path_prefix = "build" + os.path.sep if not keep_spec else ""

    try:
        scripts: Dict[str, str] = pyproject["tool"]["poetry"]["scripts"]

    except KeyError:
        print("No scripts found in pyprojects.toml")
        name = input("What is the name of the program to build? ")
        module = input("What is the runnable module to build? ")
        scripts = {name: module}

    files = []

    for name in scripts:
        path_parts = scripts[name].split(":")[0].split(".")
        package = ".".join(path_parts[0:-1])

        if not keep_spec:
            path_parts.insert(0, "..")

        win_path = r"\\".join(path_parts) + ".py"
        unix_path = "/".join(path_parts) + ".py"

        print(f"Generating {path_prefix}{name}.spec")
        with open(f"{path_prefix}{name}.spec", "w") as specfile:
            specfile.write(
                _spec_template.format(name=name, script_path_win=win_path, script_path_unix=unix_path, package=package)
            )

        files.append(f"{path_prefix}{name}.spec")

    return files
